# Converts a canonical form to a nilpotent associative algebra
InstallMethod( NAAlgByCanoForm, [IsRecord], function( C )
	return AlgebraByTable( C.cano );
end );


# Converts an nilpotent associative algebra to a canonical form
InstallMethod( CanoFormByNAAlg, [IsAlgebra and IsFiniteDimensional], function( A )
	return CanoFormWithAutGroupOfTable(NilpotentTable(A));
end );


# Build a canonical form for the d-generator F-algebra with trivial multiplication
TrivialCanoForm := function( F, d )
    local R, H, G, V, v, h, P;

    # table
    R := TrivialTable(d, F);

    # aut group
    H := GL(d,F);
    G := rec();
    G.agAutos := [];
    G.one := IdentityMat(d,F);
    G.field := F;
    G.basis := IdentityMat(d,F);
    G.partition := false;

    # add perm action
    V := F^d;
    v := Filtered(Elements(V), x -> x <> Zero(V));
    h := ActionHomomorphism(H, v, OnRight, "surjective");
    P := Image(h);
    SetSize(P, Size(H));

    # reset relevant entries in aut group record
    G.glPerms := SmallGeneratingSet(P);
    G.glAutos := List(G.glPerms, x -> PreImagesRepresentative(h,x));
    G.glOrder := Size(H);
    G.size    := Size(H);

    # add info
    return rec(cano := R, auto := G);
end;


# The operation of a matrix on a basis
OnBases := function( base, mat )
    if Length(base) = 0 then 
        return base; 
    else
        return TriangulizedMat( base * mat );
    fi;
end;


# Calculates a label for a subspace
LabelOfSubspace := function( base, info )
    local pt, j, i;

    # compute pt
    pt := List( [1..info.l], x -> 0 );
    for j in [1..info.l] do
        for i in [1..info.d] do
            if base[j][i] <> Zero(info.field) then
                pt[j] := pt[j] + (Position(info.elements, base[j][i]) - 1) * info.power[i];
            fi;
        od;
    od;

    # create label
    return pt;
end;


# Helper function for SubspaceOfLabel
CoeffsLabel := function( lab, info)
   local i, exp;
   i := info.d;
   exp := List( info.power, y -> 0 );
   while lab <> 0 do
      exp[i] := QuoInt (lab, info.power[i]);
      lab := RemInt (lab, info.power[i]);
      i := i - 1;
   od;
   return List(exp, x -> info.elements[x+1]);
end;


# Calculates a subspace from a label
SubspaceOfLabel := function ( lab, info )
   return List( [1..info.l], x -> CoeffsLabel( lab[x], info ) );
end;


# The operation of a matrix on a label
OnLabels := function( lab, mat, info )
	local base;

    if Length(lab) = 0 then 
        return lab; 
    else
		base := SubspaceOfLabel(lab, info);
		base := TriangulizedMat( base * mat );
        return LabelOfSubspace(base, info);
    fi;
end;


# Calculates all quotients by allowable subspaces that yield descendants of the given stepsize
QuotientsByAllowablesByLabels := function( C, G, s )
    local d, I, S, iter, info, all, i, r, mats, orbit, o, gen, new, reps, H, Q, des, W, M;

    d := C.mul;
    if s < 0 or d < s then return []; fi;

	Info( InfoModIsom, 1, "ByLabels: ", C.fld, " ", d, " ", d-s);

    # get all subspaces
    S := Subspaces( C.fld^d, d-s);
    Info( InfoModIsom, 1, "determined ", Size(S)," subspaces");  
    iter := Iterator(S);

	info := rec( d := d, l := d-s, field := C.fld, 
	             elements := Elements(C.fld), power := List([1..d], x -> Size(C.fld)^(x-1)));

    # filter for supplements
    I := IdentityMat(d,C.fld){[1..C.nuc]};
    all := [];

    for i in [1..Size(S)] do
       	s := BasisVectors(Basis(NextIterator(iter)));
        if C.nuc < C.mul then 
            r := RankMat(Concatenation(I, s));
            if r = d then Add( all, LabelOfSubspace(s, info) ); fi;
        else
            Add(all, LabelOfSubspace(s, info));
        fi;
    od;
    Info( InfoModIsom, 1, "reduced to ", Length(all)," supplement labels");

    # determine orbits
    M := Concatenation(List(G.agAutos, x -> x[2]), List(G.glAutos, x -> x[2]));
	reps := [];

	while Length(all) > 0 do
		orbit := [ all[1] ];
		
		for o in orbit do
			for gen in M do
				new := OnLabels(o, gen, info);

				if not new in orbit then
					Add(orbit, new);
				fi;
			od;
		od;

		Add(reps, orbit[1] );
		all := Difference(all, orbit);

		Info( InfoModIsom, 1, "found orbit of length ", Size(orbit), ". ", Length(all), " supplement labels remain");
	od;

    des := List(reps, x -> false);
    for i in [1..Length(reps)] do
        H := StructuralCopy( G );
        W := HybridMatrixCanoForm(H, SubspaceOfLabel(reps[i], info));
        Q := QuotientTableOfCover( C, W );
        InduceAutosToQuot( H, Q );
        des[i] := rec( cano := Q, auto := H );
    od;
    return des;
end;


# Calculates all quotients by allowable subspaces that yield descendants of the given stepsize
QuotientsByAllowablesBySubspaces := function( C, G, s )
    local d, I, S, iter, all, i, r, mats, os, H, Q, des, W, M;

    d := C.mul;
    if s < 0 or d < s then return []; fi;

	Info( InfoModIsom, 1, "BySubspaces: ", C.fld, " ", d, " ", d-s);

    # get all subspaces
    S := Subspaces( C.fld^d, d-s);
    Info( InfoModIsom, 1, "determined ",Size(S)," subspaces");
    iter := Iterator(S);

    # filter for supplements
    I := IdentityMat(d,C.fld){[1..C.nuc]};
    all := [];

    for i in [1..Size(S)] do
       	s := BasisVectors(Basis(NextIterator(iter)));
        if C.nuc < C.mul then 
            r := RankMat(Concatenation(I, s));
            if r = d then Add( all, s ); fi;
        else
            Add(all, s);
        fi;
    od;

    Info( InfoModIsom, 1, "reduced to ", Length(all)," supplements");

    # determine orbits
    M := Group( Concatenation(List(G.agAutos, x -> x[2]),
                              List(G.glAutos, x -> x[2])),
                G.one[2] );

    os := Orbits(M, all, OnBases);

    Info( InfoModIsom, 1, "reduced to ", Length(os)," orbits");

    des := List(os, x -> false);
    for i in [1..Length(os)] do
        H := StructuralCopy( G );
        W := HybridMatrixCanoForm(H, os[i][1]);
        Q := QuotientTableOfCover( C, W );
        InduceAutosToQuot( H, Q );
        des[i] := rec( cano := Q, auto := H );
    od;
    return des;
end;


QuotientsByAllowables := function( C, G, s )
	if CCALGS_USELABELS then
		return QuotientsByAllowablesByLabels(C, G, s);
	else
		return QuotientsByAllowablesBySubspaces(C, G, s);
	fi;
end;


# Calculates the immediate descendants of the given cano form of stepsize s
InstallMethod( ImmediateDescendantsOfCanoForm, [IsRecord, IsPosInt], function( A, s )
    local T, G, C, i;

    # catch some arguments from  
    T := StructuralCopy(A.cano);
    G := StructuralCopy(A.auto);

    # compute cover
    C := CoveringTable(T);
    
    Info(InfoModIsom, 1, "   dim(M) = ", C.mul, " and dim(N) = ", C.nuc);
    
    if C.nuc = 0 then return []; fi;
    if s = 0 then s := C.nuc; fi;

    # lift autos
    InduceAutosToMult(G, C, T);
    
    Info(InfoModIsom, 1, "   extended autos ");

    return QuotientsByAllowables(C, G, s);

end );


# Calculates the immediate descendants of the given cano form of stepsizes in slist
InstallMethod( ImmediateDescendantsOfCanoForm, [IsRecord, IsList], function( A, slist )
    local T, G, C, all, i;

    # catch some arguments from  
    T := StructuralCopy(A.cano);
    G := StructuralCopy(A.auto);

    # compute cover
    C := CoveringTable(T);
    Info(InfoModIsom, 1, "   dim(M) = ", C.mul, " and dim(N) = ", C.nuc);
    
    if C.nuc = 0 then return []; fi;
    if slist[1] = 0 then slist[1] := C.nuc; fi;

    # lift autos
    InduceAutosToMult(G, C, T);
    Info(InfoModIsom, 1, "   extended autos ");

    all := List(slist, x -> QuotientsByAllowables(C, G, x));

	return Concatenation(all);
end );


# Calculates the immediate descendants of the given cano form of all stepsizes
InstallMethod( ImmediateDescendantsOfCanoForm, [IsRecord], function( A )
    local T, G, C, all, i;

    # catch some arguments from  
    T := StructuralCopy(A.cano);
    G := StructuralCopy(A.auto);

    # compute cover
    C := CoveringTable(T);
    Info(InfoModIsom, 1, "   dim(M) = ", C.mul, " and dim(N) = ", C.nuc);
    
    if C.nuc = 0 then return []; fi;

    # lift autos
    InduceAutosToMult(G, C, T);
    Info(InfoModIsom, 1, "   extended autos ");

    all := List([1..C.mul], x -> QuotientsByAllowables(C, G, x));

	return Concatenation(all);
end );


# Calculates the immediate descendants of the given nilpotent associative algebra of stepsize s
InstallMethod( ImmediateDescendantsOfNAAlg, [IsAlgebra and IsFiniteDimensional, IsPosInt], function( A, s )
    local C, des;

	C := CanoFormByNAAlg(A);
	des := ImmediateDescendantsOfCanoForm(C, s);
	
	return List(des, x -> NAAlgByCanoForm(x));
end );


# Calculates the immediate descendants of the given nilpotent associative algebra of stepsizes in slist
InstallMethod( ImmediateDescendantsOfNAAlg, [IsAlgebra and IsFiniteDimensional, IsList], function( A, slist )
    local C, des;

	C := CanoFormByNAAlg(A);
	des := ImmediateDescendantsOfCanoForm(C, slist);
	
	return List(des, x -> NAAlgByCanoForm(x));
end );


# Calculates the immediate descendants of the given nilpotent associative algebra of all stepsizes
InstallMethod( ImmediateDescendantsOfNAAlg, [IsAlgebra and IsFiniteDimensional], function( A )
    local C, des;

	C := CanoFormByNAAlg(A);
	des := ImmediateDescendantsOfCanoForm(C);
	
	return List(des, x -> NAAlgByCanoForm(x));
end );


# Determines if the given canonical form is capable
InstallMethod( IsCapableCanoForm, [IsRecord], function( C )
    return CoveringTable(C.cano).nuc > 0; 
end );


# Determines if the given nilpotent associative algebra is capable
InstallMethod( IsCapableNAAlg, [IsAlgebra and IsFiniteDimensional], function( A )
	return IsCapableCanoForm(CanoFormByNAAlg(A));
end );


# Determines the coclass of the given canonical form
InstallMethod( CoclassOfCanoForm, [IsRecord], function( C )
    return C.cano.dim - C.cano.wgs[C.cano.dim];
end );


# Determines the coclass of the given nilpotent associative algebra
InstallMethod( CoclassOfNAAlg, [IsAlgebra and IsFiniteDimensional], function( A )
    return CoclassOfCanoForm(CanoFormByNAAlg(A));
end );


# Determines whether the given canonical form is coclass settled
InstallMethod( IsCoclassSettledCanoForm, [IsRecord], function( C )
	if 1 in List(Set(C.cano.wgs), x -> Size(Positions(C.cano.wgs, x))) then
		return true;
	elif CoveringTable(C.cano).nuc <= 1 then
		return true;
	fi;

	return false;
end );


# Determines whether the given nilpotent associative algebra is coclass settled
InstallMethod( IsCoclassSettledNAAlg, [IsAlgebra and IsFiniteDimensional], function( A )
	return IsCoclassSettledCanoForm(CanoFormByNAAlg(A));
end );


# Constructs all roots for the coclass graph G_F(r);
# arguments for this function are (field, coclass, [rank])
InstallMethod( RootsOfCoclassGraph, [IsField and IsFinite, IsInt], function ( F, r )
    local T, des, i, j, new, k, l, A, roots;

	des := List([1..2*r-2], x -> []);

	roots := [TrivialCanoForm(F,r+1)];		
	for i in [2..r] do
    	Add(des[i], TrivialCanoForm(F, i));
	od;
	
	for i in [2..2*r-2] do
		for j in [1..Length(des[i])] do
			new := ImmediateDescendantsOfCanoForm(des[i][j], [2..r - CoclassOfCanoForm(des[i][j]) + 1]);
			for k in [1..Length(new)] do
				if CoclassOfCanoForm(new[k]) = r then
					Add(roots, new[k]);
				else
					Add(des[new[k].cano.dim], new[k]);
				fi;
			od;
		od;
	od;
	
	return roots;
end );

# Constructs all e-generator roots for the coclass graph G_F(r);
# arguments for this function are (field, coclass, [rank])
InstallMethod( RootsOfCoclassGraphByRank, [IsField and IsFinite, IsInt, IsInt], function ( F, r, e )
    local T, des, i, j, new, k, l, A, roots;

	des := List([1..2*r-2], x -> []);

	if e = r+1 then
		return [TrivialCanoForm(F,r+1)];
	fi;
	roots := [];
	Add(des[e], TrivialCanoForm(F, e));
	
	for i in [2..2*r-2] do
		for j in [1..Length(des[i])] do
			new := ImmediateDescendantsOfCanoForm(des[i][j], [2..r - CoclassOfCanoForm(des[i][j]) + 1]);
			for k in [1..Length(new)] do
				if CoclassOfCanoForm(new[k]) = r then
					Add(roots, new[k]);
				else
					Add(des[new[k].cano.dim], new[k]);
				fi;
			od;
		od;
	od;
	
	return roots;
end );


# Builds the stepsize 1 descendant tree of a nilpotent associative F-algebra, i.e. the
# descendants of the same coclass.
InstallMethod( DescendantTreeOfCanoForm, [IsRecord, IsPosInt], function ( C, l )
    local des, i, j, new, k;

    # loop and construct descendants
    des := List( [1..l+1], x -> [] ); 
    des[1] := [C];

    for i in [1..l] do
        for j in [1..Length(des[i])] do
            new := ImmediateDescendantsOfCanoForm(des[i][j], 1);
			for k in [1..Length(new)] do
				new[k]!.parent := j;
			od;
            Append(des[i+1], new);
        od;
    od;
    
    return des;
end );


# Builds the stepsize 1 descendant tree of a nilpotent associative F-algebra, i.e. the
# descendants of the same coclass.
InstallMethod( DescendantTreeOfNAAlg, [IsAlgebra and IsFiniteDimensional, IsPosInt], function ( A, l )
    local C, des, i;

	C := CanoFormByNAAlg(A);

	des := DescendantTreeOfCanoForm(C, l);

	for i in [1..Length(des)] do
		des[i] := List(des[i], x -> NAAlgByCanoForm(x));
	od;
	
    return des;
end );


# Calculates canonical forms for the nilpotent associative d-generator F-algebras of dimension n
InstallMethod( CanoFormsByDimAndGens, [IsField and IsFinite, IsPosInt, IsPosInt], function( F, d, n )
    local T, alg, i, j, new, k;
    
    T := TrivialCanoForm( F, d );
    alg := List([1..n], x -> []);
    alg[d] := [T];
    for i in [d+1..n] do
        for j in [1..Length(alg[i-1])] do
            new := ImmediateDescendantsOfCanoForm(alg[i-1][j], [1..n-(i-1)]);
            for k in [1..Length(new)] do
                Add(alg[new[k].cano.dim], new[k]);
            od;
        od;
    od;
    
    return alg[n];
end );


# Calculate the nilpotent associative d-generator F-algebras of dimension n
InstallMethod( NAAlgsByDimAndGens, [IsField and IsFinite, IsPosInt, IsPosInt], function( F, d, n )
	return List(CanoFormsByDimAndGens(F, d, n), x -> NAAlgByCanoForm(x));
end );


# Calculates canonical forms for the nilpotent associative F-algebras of dimension n
InstallMethod( CanoFormsByDim, [IsField and IsFinite, IsPosInt], function( F, n )
    local algs, d;

    algs := List([1..n], x -> []);
    for d in [1..n] do
        algs[d] := CanoFormsByDimAndGens(F, d, n);
    od;

    return Flat(algs);
end );


# Calculates the nilpotent associative F-algebras of dimension n
InstallMethod( NAAlgsByDim, [IsField and IsFinite, IsPosInt], function( F, n )
	return List(CanoFormsByDim(F, n), x -> NAAlgByCanoForm(x));
end );

